#include "defs.h"
#include "jpeg.h"

#define		DEBUG			0
#define		LINEAR			1
#define		TCL_MAX			256
#define		SZVIEW			64
#define		nbmax			64
#define		szmax			256
#define		MAX_SUBSET		1000
#define		FEATURE_SIZE	128
#define		DISPLAY_SIZE	64
#define		WORK_SPACE		1000
#define		SKETCH_ID		9999999

static int parse_tcl(char *tcl_command, int *tcl_argc, char *tcl_argv[]);
static int dump_list(int subflag, int *submap, int *viewlist);
static int extract(char *db_name, struct img_set *img_set, int index, int iconsize);
static int match_template(struct img_set *img_set, struct img_set *key_set, 
						  struct img_info *curimg, int num_to_find, int *list);
static int browse(struct img_set *img_set, struct img_set *key_set, struct info_pym *info_pym, 
				  int level, int istart, int jstart, int *viewlist);
static int db_build(char *db_name, struct img_set *img_set, char *fname);
static int buildpym(struct img_set *img_set, struct img_set *key_set, 
					struct info_pym *info_pym, int pym_flag, int nbrel, int *relevance);
static int insert_img(struct img_set *img_set, struct img_set *key_set, struct img_info *curimg);

main(int argc, char *argv[]) 
{
  struct node_set best_ns;
  struct img_set  img_set, key_set, sub_set, sub_key;
  struct info_pym info_pym, sub_pym;
  char   db_name[256], longname[256], line[1024];
  float  weight[nbweight], newweight[nbweight];
  float  filter;
  float  rsize;
  int    i, j, status, index, iconsize, oldnum, oldmax;
  int    kth, nbtpl, *tpl_ix, nbrel, *rel_ix;
  int    pym_flag, subflag, *submap;
  int    level, istart, jstart, *viewlist;
  int    tcl_argc;
  int    num_class, *classsize, **semantic;
  char   *tcl_argv[TCL_MAX], **sname;
  FILE   *fp;

  for (i=0; i<nbweight; i++) weight[i] = 1.0;
  filter = 1.0;
  rsize = 1.0;
  setWeight(weight);
  setFilter(filter);
  setHeurist(rsize);

  if (argc == 1) strcpy(db_name, "db_folderSIMPLE");
  else			 strcpy(db_name, argv[1]);

  get_img_info(&img_set, db_name, "feature", WORK_SPACE, 0);
  if (img_set.num_img>0) read_img_set(&img_set, db_name, "feature", "rb");
  fprintf(stdout, "OK %d\n", img_set.num_img);
  fflush(stdout);

  get_img_info(&key_set, db_name, "SFLEXIBLE", WORK_SPACE, 1);
  if (key_set.num_img>0) read_img_set(&key_set,  db_name, "SFLEXIBLE", "rb");

  if (key_set.num_img>0) get_img_tree(&img_set, &key_set, db_name, "SFLEXIBLE");
  if (key_set.num_img>0) read_info_pym(&img_set, &key_set, &info_pym, db_name, "SFLEXIBLE");

  allocate_img_info(&sub_set, MAX_SUBSET);
  allocate_img_set(&sub_set);

  allocate_img_info(&sub_key, MAX_SUBSET / 2);
  allocate_img_set(&sub_key);

  initialize_node_set(&best_ns);
  tpl_ix   = (int *) malloc(szmax * sizeof(int));  
  rel_ix   = (int *) malloc(szmax * sizeof(int));  
  submap   = (int *) malloc(MAX_SUBSET * sizeof(int));
  viewlist = (int *) malloc(SZVIEW * sizeof(int));


  while (1) {

    fprintf(stdout,"Command> ");
    fflush(stdout);

    gets(line);

	parse_tcl(line, &tcl_argc, tcl_argv);

    if		  (tcl_argc == 0) {

	  fprintf(stdout, "\n");
	  fflush(stdout);

    } else if (!strncmp(tcl_argv[0], "filename", 10)) {

	  if (tcl_argc != 2) { fprintf(stdout, "Not enough parameter\n"); continue; }
	  index = atoi(tcl_argv[1]);
	  if (index >= 0 && index < img_set.num_img) fprintf(stdout, "OK %s\n", img_set.img[index].name);
	  else                                       fprintf(stdout, "Error\n");
	  fflush(stdout);

    } else if (!strncmp(tcl_argv[0], "locate", 10)) {

	  if (tcl_argc != 3) { fprintf(stdout, "Not enough parameter\n"); continue; }
	  subflag = atoi(tcl_argv[1]);
	  index = atoi(tcl_argv[2]);
	  if (subflag == 0) {
		if (i >= img_set.num_img) 
		  fprintf(stdout, "Error\n");
		else
		  fprintf(stdout, "OK %d %d %d\n", img_set.img[index].level, img_set.img[index].x, img_set.img[index].y);
	  } else {
		for (i=0; i<sub_set.num_img; i++) if (submap[i] == index) break;
		index = i;
		if (i >= sub_set.num_img) 
		  fprintf(stdout, "Error\n");
		else
		  fprintf(stdout, "OK %d %d %d\n", sub_set.img[index].level, sub_set.img[index].x, sub_set.img[index].y);
	  }
	  fflush(stdout);

    } else if (!strncmp(tcl_argv[0], "extract", 10)) {

	  if (tcl_argc != 3) { fprintf(stdout, "Not enough parameter\n"); continue; }
	  index = atoi(tcl_argv[1]);
	  iconsize = atoi(tcl_argv[2]);
	  extract(db_name, &img_set, index, iconsize); 
	  fprintf(stdout, "OK\n");
	  fflush(stdout);

	} else if (!strncmp(tcl_argv[0], "addlist", 10)) {

	  if (tcl_argc != 2) { fprintf(stdout, "Not enough parameter\n"); continue; }
	  if (img_set.featureblock == NULL) allocate_img_set(&img_set);
	  oldnum = img_set.num_img;
	  oldmax = img_set.max_img;
      status = db_build(db_name, &img_set, tcl_argv[1]);
	  if (status == 0) {
      write_img_info(&img_set, db_name, "feature");
	  if (img_set.num_img < MAX_SUBSET || img_set.max_img != oldmax) {
		if (key_set.num_img < img_set.num_img / 2) {
		  free_img_info(&key_set);
          allocate_img_info(&key_set, img_set.num_img / 2 + WORK_SPACE);
		  allocate_img_set(&key_set);
		}
		buildpym(&img_set, &key_set, &info_pym, 0, 0, rel_ix);
	  } else {
		  for (i=oldnum; i<img_set.num_img; i++) {
			  insert_img(&img_set, &key_set, &img_set.img[i]);
		  }
	  }
      write_img_set (&key_set, db_name, "SFLEXIBLE", "wb");
      write_img_info(&key_set, db_name, "SFLEXIBLE");
	  write_img_tree(&img_set, &key_set, db_name, "SFLEXIBLE");
      write_info_pym(&img_set, &key_set, &info_pym, db_name, "SFLEXIBLE");
	  }
	  fprintf(stdout, "OK\n");
	  fflush(stdout);

	} else if (!strncmp(tcl_argv[0], "match", 10)) {

	  if (tcl_argc != 3) { fprintf(stdout, "Not enough parameter\n"); continue; }
	  subflag = atoi(tcl_argv[1]);
	  index = atoi(tcl_argv[2]);
	  if (subflag == 0) match_template(&img_set, &key_set, &img_set.img[index], SZVIEW, viewlist);
	  else              match_template(&sub_set, &sub_key, &img_set.img[index], SZVIEW, viewlist);
	  fprintf(stdout, "OK ");
	  dump_list(subflag, submap, viewlist);
	  fprintf(stdout, "\n");
	  fflush(stdout);

	} else if (!strncmp(tcl_argv[0], "prunelist", 10)) {

	  if (tcl_argc < 2) { fprintf(stdout, "Not enough parameter\n"); continue; }
	  nbtpl = tcl_argc - 2;
	  kth = atoi(tcl_argv[1]);
	  for (i=2; i<tcl_argc; i++) tpl_ix[i-2] = atoi(tcl_argv[i]);
	  prunebylist(&img_set, &sub_set, submap, nbtpl, tpl_ix, kth);
	  fprintf(stdout, "OK%d\n", sub_set.num_img);
	  fflush(stdout);

	} else if (!strncmp(tcl_argv[0], "relevance", 10)) {

	  if (tcl_argc < 2) { fprintf(stdout, "Not enough parameter\n"); continue; }
	  subflag = atoi(tcl_argv[1]);
	  nbtpl = tcl_argc - 2;
	  for (i=2; i<tcl_argc; i++) tpl_ix[i-2] = atoi(tcl_argv[i]);

	  if (subflag == 0) {
	    relevance(&img_set, nbtpl, tpl_ix, newweight);
	  } else {
		nbrel = 0;
		for (i=0; i<nbtpl; i++) 
			for (j=0; j<sub_set.num_img; j++) 
				if (submap[j] == tpl_ix[i]) {
					rel_ix[nbrel] = j; 
					nbrel++; 
					break;
				}
	    relevance(&sub_set, nbrel, rel_ix, newweight);
	  }
      fprintf(stdout, "OK ");
	  for (i=0; i<nbweight; i++) fprintf(stdout, "%f ", newweight[i]);
	  fprintf(stdout, "\n");
	  fflush(stdout);

	} else if (!strncmp(tcl_argv[0], "buildpym", 10)) {

	  if (tcl_argc < 2) { fprintf(stdout, "Not enough parameter\n"); continue; }
	  pym_flag = atoi(tcl_argv[1]);
	  subflag = atoi(tcl_argv[2]);
	  nbtpl = tcl_argc - 3;
	  for (i=3; i<tcl_argc; i++) tpl_ix[i-3] = atoi(tcl_argv[i]);

	  if (subflag == 0) {
		if (key_set.num_img < img_set.num_img / 2) {
		  free_img_info(&key_set);
          allocate_img_info(&key_set, img_set.num_img / 2 + WORK_SPACE);
		  allocate_img_set(&key_set);
		}
	    if (pym_flag == 1) {
		  relevance(&img_set, nbtpl, tpl_ix, newweight);
  	      setWeight(newweight);
		}
		buildpym(&img_set, &key_set, &info_pym, pym_flag, 0, rel_ix);
	    write_img_set (&key_set, db_name, "SFLEXIBLE", "wb");
        write_img_info(&key_set, db_name, "SFLEXIBLE");
	    write_img_tree(&img_set, &key_set, db_name, "SFLEXIBLE");
        write_info_pym(&img_set, &key_set, &info_pym, db_name, "SFLEXIBLE");
	  } else {
		nbrel = 0;
		for (i=0; i<nbtpl; i++) 
			for (j=0; j<sub_set.num_img; j++) 
				if (submap[j] == tpl_ix[i]) {
					rel_ix[nbrel] = j; 
					nbrel++; 
					break;
				}
	    if (pym_flag == 1) {
	      relevance(&sub_set, nbrel, rel_ix, newweight);
  	      setWeight(newweight);
		}
		buildpym(&sub_set, &sub_key, &sub_pym, pym_flag, nbrel, rel_ix);
	  }
  	  setWeight(weight);

	  fprintf(stdout, "OK\n");
	  fflush(stdout);

	} else if (!strncmp(tcl_argv[0], "browse", 10)) {

	  if (tcl_argc != 5) { fprintf(stdout, "Not enough parameter\n"); continue; }
	  subflag = atoi(tcl_argv[1]);
	  level = atoi(tcl_argv[2]);
	  istart = atoi(tcl_argv[3]);
	  jstart = atoi(tcl_argv[4]);
	  if (subflag == 0) browse(&img_set, &key_set, &info_pym, level, istart, jstart, viewlist);
	  else 	            browse(&sub_set, &sub_key, &sub_pym,  level, istart, jstart, viewlist);
	  fprintf(stdout, "OK ");
	  dump_list(subflag, submap, viewlist);
	  fprintf(stdout, "\n");
	  fflush(stdout);

	} else if (!strncmp(tcl_argv[0], "exit", 10)) {
	  break;

    } else {
	  fprintf(stdout, "Unrecognized Command %s\n", tcl_argv[0]);
	}

  }

  /* free set of location nodes for each imgs */
  free_node_set(&best_ns);
  free_img_info(&sub_key);
  free_img_info(&sub_set);
  free_img_info(&key_set);
  free_img_info(&img_set);

  return(0);
}

/* Parse the relevance/irrelevance list */
static int parse_tcl(char *tcl_command, int *tcl_argc, char *tcl_argv[])
{
  char *token;
  int  i ;

  *tcl_argc = 0;
  for (i=0; i<TCL_MAX; i++) tcl_argv[i] = NULL;

  token = strtok(tcl_command, " ");
  for (i=0; i<TCL_MAX && token != NULL; i++) {
    
	if (token == NULL) return (0);
	tcl_argv[*tcl_argc] = token;
	(*tcl_argc) ++;
    token = strtok(NULL, " ");

  }
  return(0);

}

/* Parse the relevance/irrelevance list */
static int dump_list(int subflag, int *submap, int *viewlist)
{
  int  i ;

  if (subflag == 0) 
	for (i=0; i<SZVIEW; i++) fprintf(stdout, "%d ", viewlist[i]);
  else              
	for (i=0; i<SZVIEW; i++) 
	  if (viewlist[i] == SKETCH_ID) fprintf(stdout, "%d ", SKETCH_ID);
	  else if (viewlist[i] >= 0)	fprintf(stdout, "%d ", submap[viewlist[i]]);
	  else							fprintf(stdout, "-1 ");

  return(0);

}

static int extract(char *db_name, struct img_set *img_set, int index, int iconsize)
{
    struct img img_in, img_thumb;
	struct JPEG_img img_jpg;
	int    maxsize, img_type, l;
	char   longname[256];
	FILE   *fthumb;


	if (index == SKETCH_ID) {
	  if (read_img("template.ppm", &img_in, &img_type) == FAILED) return(-1);
      thumbnail(&img_in, &img_thumb, iconsize);

	} else {
      sprintf(longname, "%s/thumbnail.jpg", db_name);
      if ((fthumb = fopen(longname, "rb")) == NULL) {
        fprintf(stderr,"\nError: can't write to %s\n", longname); 
        exit(-1);
	  }

	  if (read_JPEG(fthumb, &img_jpg, img_set->img[index].thumbpt) == FAILED) return(-1);
	  maxsize = max(img_jpg.width, img_jpg.height);
	  if (iconsize == maxsize) {
		img_thumb.col   = img_jpg.width;
		img_thumb.row   = img_jpg.height;
		img_thumb.color = img_jpg.color;
		img_thumb.mono  = img_jpg.mono;
	  } else {
        img_in.col   = img_jpg.width;
        img_in.row   = img_jpg.height;
        img_in.color = img_jpg.color;
        img_in.mono  = img_jpg.mono;
		thumbnail(&img_in, &img_thumb, iconsize);
	  }


	}

	write_img("extract.ppm", &img_thumb, PPM_IMAGE);
	for (l=0; l<3; l++) free_img((void **) img_thumb.color[l]);
	free((void *) img_thumb.color);
	fclose(fthumb);

	return(0);
}


static int browse(struct img_set *img_set, struct img_set *key_set, struct info_pym *info_pym, 
				  int level, int istart, int jstart, int *viewlist)
{
	int img_id, found, i, j, k, index;

	index = 0;
	for (i=0; i<8; i++) 
	for (j=0; j<8; j++) {
		found = 0;
		for (k=0; k<key_set->num_img && found == 0; k++) {
			if (key_set->img[k].level == level &&
				key_set->img[k].x == istart + i &&
				key_set->img[k].y == jstart + j) {
				found = 1;
				img_id = key_set->img[k].nearest->ix;
			}
		}
		for (k=0; k<img_set->num_img && found == 0; k++) {
			if (img_set->img[k].level == level &&
				img_set->img[k].x == istart + i &&
				img_set->img[k].y == jstart + j) {
				found = 1;
				img_id = k;
			}
		}
		if (found == 0) viewlist[index++] = -1;
		else			viewlist[index++] = img_id;
	}
    return(0);
}


static int match_template(struct img_set *img_set, struct img_set *key_set, 
						  struct img_info *curimg, int num_to_find, int *viewlist)
{
    struct node_set best_ns;
	struct node p;
	int    i;
	double dist;


    initialize_node_set(&best_ns);

    if (LINEAR == 1 || key_set->num_img == 0) {
        fast_db_search(img_set, key_set, &best_ns, curimg, num_to_find); 
	} else {
        list_db_search(img_set, key_set, img_set->root, &best_ns, curimg, num_to_find);
	}

    /* print out info about three best results */
    for (i=0; i<num_to_find && i<img_set->num_img; i++) {
        if (get_best_from_node_set(&best_ns,&p,&dist) == -1) break;
        viewlist[i] = p.k;
	}
	for (;i<num_to_find; i++) viewlist[i] = -1;

	free_node_set(&best_ns);
    return(0);

}


static int insert_img(struct img_set *img_set, struct img_set *key_set, struct img_info *curimg) 
{
  struct img_info *nearimg, *preimg, *chkimg, *curkey, *newkey, *map[2][2];
  struct node_set best_ns;
  struct node p;
  double dist;
  int    i, j, di, dj;

    initialize_node_set(&best_ns);

    if (LINEAR == 1 || key_set->num_img == 0) {
        fast_db_search(img_set, key_set, &best_ns, curimg, 1); 
	} else {
        list_db_search(img_set, key_set, img_set->root, &best_ns, curimg, 1);
	}

    /* print out info about three best results */
    get_best_from_node_set(&best_ns,&p,&dist);

	free_node_set(&best_ns);

	nearimg = &img_set->img[p.k];
	curkey = nearimg->key;
	for (i=0; i<2; i++) for (j=0; j<2; j++) map[i][j] = NULL;
	preimg = NULL;
	for (chkimg=curkey->child, i=0; chkimg!=NULL; chkimg=chkimg->next, i++) {
		map[chkimg->y][chkimg->x] = chkimg;
		if (chkimg->next == nearimg) preimg = chkimg;
	}

	if (i >= 4) { /* Split */

		append_img_info(key_set, curimg);
		newkey = &key_set->img[key_set->num_img-1];

		newkey->nearest = nearimg;
		newkey->size = 0;
		newkey->dist = 0;

		newkey->key = curkey;
		newkey->next = nearimg->next;
		if (preimg == NULL) curkey->child = newkey;
		else                preimg->next  = newkey;

		newkey->child = curimg;
		curimg->next  = nearimg;
		nearimg->next = NULL;
		curimg->key = newkey;
		nearimg->key = newkey;

		newkey->level = nearimg->level;
		newkey->x   = nearimg->x;
		newkey->y   = nearimg->y;

		nearimg->level = newkey->level + 1;
		nearimg->x   = newkey->x * 2;
		nearimg->y   = newkey->y * 2;

		curimg->level = newkey->level + 1;
		curimg->x   = newkey->x * 2 + 1;
		curimg->y   = newkey->y * 2;

	} else { /* Insert */

		for (i=0; i<2; i++)	for (j=0; j<2; j++) if (map[i][j] == NULL) {dj = j; di = i; }

		curimg->key   = curkey;
		curimg->next  = curkey->child;
		curkey->child = curimg;

		curimg->level = curkey->level + 1;
		curimg->y   = curkey->y * 2 + di;
		curimg->x   = curkey->x * 2 + dj;

	}

    return(0);

}


static int buildpym(struct img_set *img_set, struct img_set *key_set, 
					struct info_pym *info_pym, int pym_flag, int nbrel, int *relevance) 
{
  struct img_set work_set;
  struct img_info *curimg, *root;
  struct feature dfeature;

  allocate_img_info(&work_set, img_set->num_img + WORK_SPACE);

  if (DEBUG) fprintf(stdout, "Building Binary Tree\n");
  if (pym_flag == 2) {
    if (img_set->num_img > MAX_SUBSET) buildtree1(img_set, key_set, &work_set);
    else                               buildtree3(img_set, &work_set, nbrel, relevance);
  } else {
    if (img_set->num_img > MAX_SUBSET) buildtree1(img_set, key_set, &work_set);
    else                               buildtree2(img_set, &work_set);
  }
  
  /* code from regroup.c */
  root = img_set->root;
  root->next = NULL;
  root->key  = NULL;
  root->feature = &dfeature;
  reset_feature(root->feature);
  key_set->num_img = 0;
  if (DEBUG) fprintf(stdout, "Converting to Quad-Tree\n");
  regroup(img_set, &work_set, key_set, root, 0, nbrel, relevance);

  if (root->type == 1) img_set->root = key_set->root = root->child;
  for (curimg=img_set->root; curimg!=NULL; curimg=curimg->next) {
    curimg->key = NULL;
    curimg->dist = MAXVALUE;
  }

  /* code from mapping.c */
  if (DEBUG) fprintf(stdout, "Mapping to Pyramid\n");
  mapping(img_set);

  if (DEBUG) fprintf(stdout, "End of Build pyramid\n");

  free_img_info(&work_set);
  return(0);

}


static int db_build(char *db_name, struct img_set *img_set, char *fname) 
{
  struct  img       img_in, img_thumb, img_display;
  struct  float_img fimg;
  struct  img_info	img_info, *curimg;
  struct  JPEG_img img_out;
  char    buffer[80], longname[256];
  int     k, l, num_img, img_type;
  FILE    *flist, *ffeature, *fthumb;

  if ((flist = fopen(fname, "r")) == NULL) return(-1);
  
  if (fscanf(flist, "%d\n", &num_img) != 1) return(-1);

  if (num_img <= 0) return(-1);

  ffeature = open_database(db_name, "feature", "ab");
  sprintf(longname, "%s/thumbnail.jpg", db_name);
  if ((fthumb = fopen(longname, "ab")) == NULL) {
    fprintf(stderr,"\nError: can't write to %s\n", longname); 
    exit(-1);
  }

  fseek(fthumb, 0, SEEK_END);
  /* get the name of each image in the database */
  /* process each image in the database */
  for(k=0; k<num_img; k++) {
    if (fgets(buffer, 80, flist) == NULL) return(-1);
    buffer[strlen(buffer)-1] = '\0';

    /* Read input image */
    if (read_img(buffer, &img_in, &img_type)==FAILED) {
      fprintf(stdout,"\nWarning: Can't read image %s\n", buffer);
      return(-1);
    }

    strcpy(img_info.name, buffer);
    img_info.ix   = k;
    img_info.type = 0;
    img_info.row  = img_in.row;
    img_info.col  = img_in.col;
    img_info.num_img = 0;

    /* Subsample to 128x128 image and Do color space transform */
    mono_to_color(&img_in, img_type);
    thumbnail(&img_in, &img_thumb, FEATURE_SIZE);
    color_space_transform(&img_thumb, &fimg);

    thumbnail(&img_thumb, &img_display, DISPLAY_SIZE);
    img_out.jpeg_type = 'c';
    img_out.width     = img_display.col;
    img_out.height    = img_display.row;
    img_out.color     = img_display.color;
    img_out.mono      = img_display.mono;

	img_info.thumbpt = ftell(fthumb);
    if (write_JPEG(fthumb, &img_out, 85, 0) == FAILED) { /* quality=90, smooth=0 */
      fprintf(stderr,"\nError: can't write to JPEG image %s\n", longname); 
      fclose(fthumb);
      exit(-1);
	}
    append_img_info(img_set, &img_info);
	curimg = &img_set->img[img_set->num_img-1];

    /* Build pyramid of histogram */
    build_feature(&fimg, curimg->feature);
    if (write_feature(curimg->feature, ffeature, &(curimg->featurept), &(curimg->featurelen)) == FAILED) {
      fprintf(stdout,"\nError: Can not write to image database\n");
	  exit(-1);
    }

    for (l=0; l<3; l++) free_img((void **) img_display.color[l]);
    free( (void *) img_display.color );

  }
  fclose(fthumb);
  fclose(ffeature);
  fclose(flist);
  return(0);

}


