#include "defs.h"

#define DEBUG			0
#define COMBINATIONNO	8
#define PI				3.141592653589793
#define SPARSE			20
#define MAXSPARSE		100

struct snode {
		float dist;
        int   row, col, count;
        struct snode *nextrow;
        struct snode *nextcol;
} *disttbl;


/* routine for agglomerate clustering */
static struct img_info *agglomerate1(struct img_set  *img_set, struct img_set *key_set);
static void insert(struct snode *curdist, int row, int col);
static void detach(struct snode *curdist);
static float cost (float dist1, int num1, float dist2, int num2, float dist3,int num3);
static float guess(float maxsize1, float minsize1, float sumsize1, int num1, 
			       float maxsize2, float minsize2, float sumsize2, int num2);



int tblix = 0;
struct snode *rowhead, *colhead;
float  *maxsize, *minsize, *sumsize, meansize, lowersize, uppersize;
float  *mindist;
int    *mincluster;
struct img_info **clusterptr;

buildtree1(struct img_set *img_set, struct img_set *search_set, struct img_set *binary_set) 
{
  struct node_set best_ns;
  struct img_info *curimg;
  struct node p;
  struct snode *curdist;
  int    row, col;
  double dist;
  int    num_to_find, size, found;
  int    i, j;


  /* Initialize set of location nodes for each image */
  initialize_node_set(&best_ns);

  if (DEBUG) fprintf(stdout, "Begin Partitioning\n");

  search_set->num_img = 0;
  for (i=0; i<img_set->num_img; i++) {
    if (i == img_set->num_img-1) img_set->img[i].next  = NULL;
	else						 img_set->img[i].next  = &(img_set->img[i+1]);
    img_set->img[i].child = NULL;
    img_set->img[i].key   = NULL;
    img_set->img[i].nearest  = NULL;
  }
  search_set->root=img_set->root=partition(img_set, search_set, img_set->img, 0);

  /* code from proximity.c */
  if (DEBUG) fprintf(stdout, "Compute/Read sparse matrix\n");
  num_to_find = min(img_set->num_img, SPARSE);
  size = img_set->num_img;

  rowhead = (struct snode *) malloc(sizeof(struct snode) * size);
  colhead = (struct snode *) malloc(sizeof(struct snode) * size);
  /* Initial empty sparse matrix */
  for (i=0; i<img_set->num_img; i++) {
    colhead[i].nextrow = &(colhead[i]); 
    colhead[i].nextcol = &(colhead[i]);
    colhead[i].col     = i;
    colhead[i].row     = -1;
	colhead[i].count   = 0;
    rowhead[i].nextcol = &(rowhead[i]); 
    rowhead[i].nextrow = &(rowhead[i]);
    rowhead[i].col     = -1;
    rowhead[i].row     = i;
	rowhead[i].count   = 0;
  }

  disttbl = (struct snode *) malloc(sizeof(struct snode) * size * (num_to_find+1));
  clusterptr  = (struct img_info **)  malloc(sizeof(struct img_info *) * 
                                             img_set->num_img);
  mincluster  = (int *)   malloc(sizeof(int)   * img_set->num_img);
  mindist     = (float *) malloc(sizeof(float) * img_set->num_img);
  maxsize     = (float *) malloc(sizeof(float) * img_set->num_img);
  minsize     = (float *) malloc(sizeof(float) * img_set->num_img);
  sumsize     = (float *) malloc(sizeof(float) * img_set->num_img);

  tblix = 0;
  lowersize = MAXVALUE;
  uppersize = 0;
  meansize = 0;
  setHeurist((float) 0.1);
  /* MAIN LOOP */
  for (i=0; i<img_set->num_img; i++) {
    curimg = &(img_set->img[i]);

    /* Perform fast search of database */
    reset_node_set(&best_ns);

    if (search_set->num_img == 0) {
      fast_db_search(img_set,search_set,&best_ns,curimg, num_to_find); 
    } else {
      list_db_search(img_set,search_set,img_set->root,&best_ns,curimg, num_to_find);
    }

    maxsize[i] = 0;
    found = 0;
    for (j=0; j<num_to_find; j++) {
      if (get_best_from_node_set(&best_ns, &p, &dist) == -1) break;
      if (dist == MAXVALUE) {
        fprintf(stdout, "Error distance %d %d %g\n", j, p.k, dist);
        exit(0);
      }
      if (DEBUG == 2) fprintf(stdout, "<%d %g>", p.k, dist);
      if (p.k == i) found = 1;

      col = max(i, p.k);
      row = min(i, p.k);
      curdist   = &(disttbl[tblix]);
      curdist->dist = dist;
      if (maxsize[i] < curdist->dist) maxsize[i] = curdist->dist;
      insert(curdist, row, col);
      tblix++;
    }
    if (found == 0) {
      col = row = i;
      curdist   = &(disttbl[tblix]);
      curdist->dist = 0;
      insert(curdist, row, col);
      tblix++;
    }
    minsize[i] = maxsize[i];
    sumsize[i] = maxsize[i];
    meansize += maxsize[i];
    if (lowersize > maxsize[i]) lowersize = maxsize[i];
    if (uppersize < maxsize[i]) uppersize = maxsize[i];
  }
  setHeurist((float) 1.0);
  meansize /= img_set->num_img;

  search_set->num_img = 0;


  if (DEBUG) fprintf(stdout, "Initialize mincluster etc\n");
  /* Initialize the cluster type/index, mindist, mincluster table */
  for (i=0; i<img_set->num_img; i++) {
    clusterptr[i] = &(img_set->img[i]);
    clusterptr[i]->size = maxsize[i];

    mindist[i] = MAXVALUE;
    mincluster[i] = i;

    curdist = colhead[i].nextrow; 
    while (curdist != &colhead[i] && curdist != &rowhead[i]) {
      dist = curdist->dist;

      if (mindist[i] > dist && curdist->row != curdist->col) {
        mindist[i] = dist;
        if (curdist->row < i) mincluster[i] = curdist->row;
        else                  mincluster[i] = curdist->col;
      }
      if (curdist->row < i) curdist = curdist->nextrow; 
      else                  curdist = curdist->nextcol; 
    }
  }
      

  for (i=0; i<img_set->num_img; i++) {
    img_set->img[i].num_img = 1;
    img_set->img[i].next  = NULL;
    img_set->img[i].child = NULL;
    img_set->img[i].key   = NULL;
    img_set->img[i].nearest  = NULL;
  }

  if (DEBUG) fprintf(stdout, "Begin Clustering\n");

  binary_set->root=img_set->root=agglomerate1(img_set, binary_set);

  for (i=0; i<img_set->num_img; i++) {
    img_set->img[i].size = 0;
  }
  /* Free proximity matrix */

  if (DEBUG) fprintf(stdout, "End Clustering\n");

  free(sumsize);
  free(minsize);
  free(maxsize);
  free(mindist);
  free(mincluster);
  free(disttbl);
  free(colhead);
  free(rowhead);
  
  return(0);

}


static struct img_info *agglomerate1(struct img_set *img_set, struct img_set *key_set) 
{
  struct img_info *curkey, *nextkey, *newkey;
  struct snode *curdist1, *curdist2, *curdist, *chkdist;
  struct snode *nextdist1, *nextdist2;
  float  dist, guessdist;
  int    currow1, currow2;
  int    i, j, ii, jj, m, n;

  newkey = &(img_set->img[0]);
  for (i=img_set->num_img; i>1; i--) {

    if (DEBUG) fprintf(stdout, "Clustering %d\n", i);

    /* Find closest clusters */ 
    dist = MAXVALUE;
    ii = jj = 0;
    for (j=0; j<img_set->num_img; j++) {
      if (dist > mindist[j]) {
        ii = j; jj = mincluster[j]; dist = mindist[j];
      }
    }

    if (DEBUG) 
      fprintf(stdout, "Closest cluster %d %d dist %g\n", ii, jj, mindist[ii]);

    if (ii == 0 && jj == 0) {
	  for (ii=0;    ii<img_set->num_img; ii++) if (clusterptr[ii] != NULL) break;
	  for (jj=ii+1; jj<img_set->num_img; jj++) if (clusterptr[jj] != NULL) break;
	} else if (ii >= jj) { /* This should not happen */
      fprintf(stdout, "Error: ii (%d) >= jj (%d) This should not happen (%dth)\n", ii, jj, i);
	  for (ii=0;    ii<img_set->num_img; ii++) if (clusterptr[ii] != NULL) break;
	  for (jj=ii+1; jj<img_set->num_img; jj++) if (clusterptr[jj] != NULL) break;
    }

    /* create a new cluster with children ii, jj */
    curkey = clusterptr[ii];
    nextkey = clusterptr[jj];
    append_img_info(key_set, curkey);
    newkey = &(key_set->img[key_set->num_img-1]);
    strcpy(newkey->name, "KEY");

    newkey->type      = 1;
    newkey->num_img   = curkey->num_img + nextkey->num_img;
    newkey->dist      = i;
    newkey->size      = mindist[ii];
    newkey->nearest   = NULL;
    newkey->key       = NULL;
    newkey->next      = NULL;
    newkey->child     = curkey;
    curkey->next      = nextkey;
    curkey->key       = newkey;
    nextkey->key      = newkey;

    if (DEBUG) fprintf(stdout, "Updating Proximity matrix\n");
    /* Update Proximity matrix, clusterptr */
    /* Merge cluster ii and jj */
    if (DEBUG) fprintf(stdout, "Merge cluster ii, jj\n");

    curdist1=colhead[ii].nextrow; curdist2=colhead[jj].nextrow;
    dist = 0;
    while ((curdist1 != &colhead[ii] && curdist1 != &rowhead[ii]) ||
           (curdist2 != &colhead[jj] && curdist2 != &rowhead[jj])) {
      /* initialize */
      currow1 = currow2 = img_set->num_img;
      nextdist1 = &colhead[ii];
      nextdist2 = &colhead[jj];
      if (curdist1 != &colhead[ii] && curdist1 != &rowhead[ii]) { 
        currow1   = curdist1->row;
        if (curdist1->row < ii) {
          nextdist1 = curdist1->nextrow; 
          currow1   = curdist1->row;
        } else {
          nextdist1 = curdist1->nextcol; 
          currow1   = curdist1->col;
        }
      }
      if (curdist2 != &colhead[jj] && curdist2 != &rowhead[jj]) {
        currow2   = curdist2->row;
        if (curdist2->row < jj) {
          nextdist2 = curdist2->nextrow; 
          currow2   = curdist2->row;
        } else {
          nextdist2 = curdist2->nextcol; 
          currow2   = curdist2->col;
        }
      }
      
      /* both distance exist */
      if (currow1 == currow2) {
        if (DEBUG == 2) fprintf(stdout, "%d %d %d %d\n", 
                   curdist1->row, curdist1->col, curdist2->row, curdist2->col);
        curdist1->dist = cost(curdist1->dist, clusterptr[ii]->num_img,  
                              curdist2->dist, clusterptr[jj]->num_img, 
                              mindist[ii],    clusterptr[currow1]->num_img);

        if (currow1 == ii) curdist1->dist = 0;
        if (dist < curdist1->dist) dist = curdist1->dist;

        if (curdist2 == nextdist1) nextdist1 = nextdist1->nextcol; 
        detach(curdist2);

        if (mindist[currow1] > curdist1->dist && currow1 != ii) {
          mindist[currow1] = dist;
          mincluster[currow1] = ii;
        }

        curdist1 = nextdist1;
        curdist2 = nextdist2;

      /* Only cluster ii */
      } else if (currow1 < currow2) {
        if (DEBUG == 2) fprintf(stdout, "%d %d %d %d\n", 
			                   curdist1->row, curdist1->col, curdist2->row, curdist2->col);
        m = currow1; n = jj;
        guessdist = 
          guess(maxsize[m], minsize[m], sumsize[m], clusterptr[m]->num_img,
                maxsize[n], minsize[n], sumsize[n], clusterptr[n]->num_img);
        guessdist = max(guessdist, mindist[ii]);
        curdist1->dist = cost(curdist1->dist, clusterptr[ii]->num_img,  
                              guessdist,      clusterptr[jj]->num_img, 
                              mindist[ii],    clusterptr[currow1]->num_img);
        if (dist < curdist1->dist) dist = curdist1->dist;

        if (mindist[currow1] > curdist1->dist && currow1 != ii) {
          mindist[currow1] = dist;
          mincluster[currow1] = ii;
        }

        curdist1 = nextdist1;

      /* Only cluster jj */
      } else if (currow1 > currow2) {
        if (DEBUG == 2) fprintf(stdout, "%d %d %d %d\n", 
			                   curdist1->row, curdist1->col, curdist2->row, curdist2->col);
        m = ii; n = currow2;
        guessdist = 
          guess(maxsize[m], minsize[m], sumsize[m], clusterptr[m]->num_img,
                maxsize[n], minsize[n], sumsize[n], clusterptr[n]->num_img);
        guessdist = max(guessdist, mindist[ii]);
        curdist2->dist = cost(guessdist,      clusterptr[ii]->num_img,  
                              curdist2->dist, clusterptr[jj]->num_img, 
                              mindist[ii],    clusterptr[currow2]->num_img);
        if (dist < curdist2->dist) dist = curdist2->dist;

        detach(curdist2);
        if (currow2 != jj &&
			(colhead[currow2].count < MAXSPARSE || colhead[ii].count < MAXSPARSE ||
			 mincluster[currow2] == jj)) {
          if (currow2 < ii) insert(curdist2, currow2, ii);
          else              insert(curdist2, ii, currow2);
          if (mindist[currow2] > curdist2->dist && currow2 != ii) {
            mindist[currow2] = dist;
            mincluster[currow2] = ii;
          }

        }

        curdist2 = nextdist2;
      }
        
    }

    maxsize[ii] = max(maxsize[ii], maxsize[jj]);
    minsize[ii] = min(minsize[ii], minsize[jj]);
    sumsize[ii] = sumsize[ii] + sumsize[jj];
    clusterptr[ii] = newkey;
    clusterptr[jj] = NULL;
    mincluster[jj] = -1;
    mindist[jj] = MAXVALUE;


    /* Update mindist/mincluster table */      
    if (DEBUG) fprintf(stdout, "Updating mindist/mincluster\n");

    chkdist = colhead[ii].nextrow; 
    while (chkdist != &(colhead[ii]) && chkdist != &(rowhead[ii])) {
      if (chkdist->row < ii) j = chkdist->row;
      else                   j = chkdist->col;
      if (mincluster[j] == ii || mincluster[j] == jj) {
        mindist[j] = MAXVALUE;
        mincluster[j] = 0;
        curdist = colhead[j].nextrow; 
        while (curdist != &(colhead[j]) && curdist != &(rowhead[j])) {
          dist = curdist->dist;
          if (mindist[j] > dist && curdist->row != curdist->col) {
            mindist[j] = dist;
            if (curdist->row < j) mincluster[j] = curdist->row;
            else                  mincluster[j] = curdist->col;
          }
          if (curdist->row < j) curdist = curdist->nextrow; 
          else                  curdist = curdist->nextcol; 
        }
      }
      if (chkdist->row < ii) chkdist = chkdist->nextrow; 
      else                   chkdist = chkdist->nextcol; 
    }

  }

  return(newkey);

}

static void insert(struct snode *curdist, int row, int col)
{
  struct snode *curcol, *prvcol;
  struct snode *currow, *prvrow;

  curcol = colhead[col].nextrow; 
  prvcol =&colhead[col]; 
  while (curcol != &(colhead[col]) && curcol->row < row) {
    prvcol = curcol;
    curcol = curcol->nextrow;
  }
  if (curcol->row == row) return;

  currow = rowhead[row].nextcol; 
  prvrow =&rowhead[row]; 
  while (currow != &(rowhead[row]) && currow->col < col) {
    prvrow = currow;
    currow = currow->nextcol;
  }
  if (currow->col == col) return;

  colhead[col].count++;
  if (col != row) colhead[row].count++;
  curdist->row = row;
  curdist->col = col;
 
  curdist->nextrow = curcol;
  prvcol->nextrow  = curdist;
  
  curdist->nextcol = currow;
  prvrow->nextcol  = curdist;
  
}

static void detach(struct snode *curdist) 
{
  struct snode *curcol, *prvcol;
  struct snode *currow, *prvrow;

  curcol = colhead[curdist->col].nextrow; 
  prvcol =&colhead[curdist->col]; 
  while (curcol != &(colhead[curdist->col]) && curcol->row < curdist->row) {
    prvcol = curcol;
    curcol = curcol->nextrow;
  }
  if (curcol->row != curdist->row) return;

  currow = rowhead[curdist->row].nextcol; 
  prvrow =&rowhead[curdist->row]; 
  while (currow != &(rowhead[curdist->row]) && currow->col < curdist->col) {
    prvrow = currow;
    currow = currow->nextcol;
  }
  if (currow->col != curdist->col) return;

  colhead[curdist->col].count--;
  if (curdist->col != curdist->row) colhead[curdist->row].count--;
  prvcol->nextrow  = curdist->nextrow;
  prvrow->nextcol  = curdist->nextcol;

  curdist->nextrow = NULL;
  curdist->nextcol = NULL;

}

static float cost(float dist1, int num1, float dist2, int num2, float dist3, int num3)
{
  float dist;

  /* Flexible - Dilating (beta=-1) */
  /* 
    dist = dist1 * (1 - BETA) * 0.5 + dist2 * (1 - BETA) * 0.5 + dist3 * BETA;
  */
  dist = dist1 + dist2 - dist3;

  return(dist);

}

static float guess(float maxsize1, float minsize1, float sumsize1, int num1, 
            float maxsize2, float minsize2, float sumsize2, int num2)
{
  float dist;

  /* Flexible - Dilating (beta=-1) */
  dist = max(sumsize1 * num2,  sumsize2 * num1); 

  return(dist);

}


