#include "defs.h"

#define DEBUG			0

#define col(a,b)	((a) > (b) ? (a-b) : (b-a))
#define row(a,b)	((a) < (b) ? (a) : (b))


/* routine for agglomerate clustering */
static struct img_info *agglomerate2(struct img_set  *img_set, struct img_set *key_set);
static struct img_info *agglomerate3(struct img_set  *img_set, struct img_set *key_set);



float  **distbuf;

float  *mindist;
int    *mincluster, *relcluster;
struct img_info **clusterptr;

int buildtree2(struct img_set *img_set, struct img_set *key_set) 
{
  struct node_set best_ns;
  struct img_info *curimg, *chkimg;
  double dist;
  int    i, j;


  /* code from proximity.c */
  distbuf = (float **) malloc(sizeof(float *) * img_set->num_img);
  for (i=0; i<img_set->num_img; i++) {
    distbuf[i] = (float *) malloc(sizeof(float) * (img_set->num_img - i));
  }
  initialize_node_set(&best_ns);

  /* MAIN LOOP */
  for (j=0; j<img_set->num_img; j++) {
    curimg = &(img_set->img[j]);

    for (i=j; i<img_set->num_img; i++) {
      chkimg = &(img_set->img[i]);
      distbuf[j][i-j] = direct_search(chkimg, curimg, &best_ns);
    }

  }


  /* Initialize the cluster type/index, mindist, mincluster table */
  clusterptr  = (struct img_info **)  malloc(sizeof(struct img_info *) * 
                                             img_set->num_img);
  mincluster  = (int *)   malloc(sizeof(int)   * img_set->num_img);
  mindist     = (float *) malloc(sizeof(float) * img_set->num_img);
  for (i=0; i<img_set->num_img; i++) {
    clusterptr[i] = &(img_set->img[i]);

    mindist[i] = MAXVALUE;
    mincluster[i] = i;
    for (j=0; j<img_set->num_img; j++) {
      dist = distbuf[row(i, j)][col(i, j)];
      if (mindist[i] > dist && i != j) {
        mindist[i] = dist;
        mincluster[i] = j;
      }
    }
  }
      

  for (i=0; i<img_set->num_img; i++) {
    img_set->img[i].num_img = 1;
    img_set->img[i].next  = NULL;
    img_set->img[i].child = NULL;
    img_set->img[i].key   = NULL;
    img_set->img[i].nearest = NULL;
  }

  if (DEBUG) fprintf(stdout, "Begin Clustering\n");

  key_set->num_img = 0;
  key_set->root=img_set->root=agglomerate2(img_set, key_set);

  for (i=0; i<img_set->num_img; i++) {
    img_set->img[i].size = 0;
  }

  /* Free proximity matrix */
  for (i=0; i<img_set->num_img; i++) free(distbuf[i]);
  free(distbuf);


  return(0);

}

static struct img_info *agglomerate2(struct img_set *img_set, struct img_set *key_set) 
{
  struct img_info *curkey, *nextkey, *newkey;
  float  dist, dist1, dist2, dist3;
  int    num1, num2;
  int    i, j, k, ii, jj;
  int    maxcluster, newmax;

  maxcluster = img_set->num_img;
  for (i=img_set->num_img; i>1; i--) {

	if (DEBUG) fprintf(stdout, "Cluster %d\n", i);

    /* Find closest clusters */ 
    dist = MAXVALUE;
    ii = jj = 0;
    for (j=0; j<maxcluster; j++) {
      if (dist > mindist[j]) {
        ii = j; jj = mincluster[j]; dist = mindist[j];
      } else if (dist == mindist[j]) {
        if (clusterptr[ii]->num_img + clusterptr[jj]->num_img >
            clusterptr[j]->num_img + clusterptr[mincluster[j]]->num_img) {
          ii = j; jj = mincluster[j]; dist = mindist[j];
        }
      }
    }

    if (DEBUG) fprintf(stdout, "Closest cluster %d %d dist %g\n", ii, jj, mindist[ii]);

    /* create a new cluster with children ii, jj */
    curkey = clusterptr[ii];
    nextkey = clusterptr[jj];
    append_img_info(key_set, curkey);
    newkey = &(key_set->img[key_set->num_img-1]);
    strcpy(newkey->name, "KEY");

    newkey->type      = 1;
    newkey->num_img   = curkey->num_img + nextkey->num_img;
    newkey->dist      = i;
    newkey->size      = mindist[ii];
    newkey->nearest   = NULL;
    newkey->key       = NULL;
    newkey->next      = NULL;
    newkey->child     = curkey;
    curkey->next      = nextkey;
    curkey->key       = newkey;
    nextkey->key      = newkey;

    if (DEBUG) fprintf(stdout, "Updating Proximity matrix\n");
    /* Update Proximity matrix, clusterptr */
    /* replace ii */

    num1  = clusterptr[ii]->num_img; 
    num2  = clusterptr[jj]->num_img; 
    dist3 = newkey->size;
    for (j=0; j<maxcluster; j++) {
      if (clusterptr[j] == NULL) continue;
      dist1 = distbuf[row(ii, j)][col(ii, j)];
      dist2 = distbuf[row(jj, j)][col(jj, j)];

      /* Flexible - Dilating (beta=-1) */
      distbuf[row(ii, j)][col(ii, j)] = dist1 + dist2 - newkey->size;
    }
 
    /* delete jj */
    if (DEBUG) fprintf(stdout, "Delete cluster jj\n");
    clusterptr[ii] = newkey;
    clusterptr[jj] = NULL;
    mincluster[jj] = -1;
    mindist[jj] = MAXVALUE;
    for (j=0; j<maxcluster; j++) distbuf[row(jj, j)][col(jj, j)] = MAXVALUE;
    for (newmax=maxcluster-1; newmax>=0; newmax--)
      if (clusterptr[newmax] != NULL) break;
    maxcluster = newmax + 1;


    distbuf[row(ii, ii)][col(ii, ii)] = 0;
    distbuf[row(jj, jj)][col(jj, jj)] = 0;

    /* Update mindist/mincluster table */      
    if (DEBUG) fprintf(stdout, "Updating mindist/mincluster\n");
    for (j=0; j<maxcluster; j++) {
      if (clusterptr[j] == NULL) continue;
      if (mincluster[j] == ii || mincluster[j] == jj) {
        num1 = clusterptr[j]->num_img;
        mindist[j] = MAXVALUE;
        mincluster[j] = 0;
        for (k=0; k<maxcluster; k++) {
          if (clusterptr[k] == NULL) continue;
          dist = distbuf[row(j, k)][col(j, k)];
          if (mindist[j] > dist && k != j) {
            mindist[j] = dist;
            mincluster[j] = k;
          } else if (mindist[j] == dist && k != j) {
            if (clusterptr[k]->num_img < clusterptr[mincluster[j]]->num_img) {
              mindist[j] = dist;
              mincluster[j] = k;
            }
          }
        }
      }
    }

  }

  return(newkey);

}

int buildtree3(struct img_set *img_set, struct img_set *key_set, int nbrel, int *relevance) 
{
  struct node_set best_ns;
  struct img_info *curimg, *chkimg;
  double dist;
  int    i, j;
  float  *disttbl;

  /* Check for correct number of arguements */

  /* code from proximity.c */
  disttbl = (float *) malloc(sizeof(float) * img_set->num_img);
  distbuf = (float **) malloc(sizeof(float *) * img_set->num_img);
  for (i=0; i<img_set->num_img; i++) {
    distbuf[i] = (float *) malloc(sizeof(float) * (img_set->num_img - i));
  }
  initialize_node_set(&best_ns);

  /* MAIN LOOP */
  disttbl = (float *) malloc(sizeof(float) * img_set->num_img);
  for (j=0; j<img_set->num_img; j++) {
    curimg = &(img_set->img[j]);
 
    disttbl[j] = MAXVALUE;
    for (i=0; i<nbrel; i++) {
      chkimg = &(img_set->img[relevance[i]]);
      dist = direct_search(chkimg, curimg, &best_ns);
      if (disttbl[j] > dist) disttbl[j] = dist;
    }

  }

  for (j=0; j<img_set->num_img; j++) {
    curimg = &(img_set->img[j]);
 
    distbuf[j][0] = 0;
    for (i=j+1; i<img_set->num_img; i++) {
      chkimg = &(img_set->img[i]);
      distbuf[j][i-j] = direct_search(chkimg, curimg, &best_ns);
      if (distbuf[j][i-j] > disttbl[i] + disttbl[j])
        distbuf[j][i-j] = disttbl[i] + disttbl[j];
    }

  }


  /* Initialize the cluster type/index, mindist, mincluster table */
  clusterptr  = (struct img_info **)  malloc(sizeof(struct img_info *) * 
                                             img_set->num_img);
  relcluster  = (int *)   malloc(sizeof(int)   * img_set->num_img);
  mincluster  = (int *)   malloc(sizeof(int)   * img_set->num_img);
  mindist     = (float *) malloc(sizeof(float) * img_set->num_img);
  for (i=0; i<img_set->num_img; i++) {
    clusterptr[i] = &(img_set->img[i]);

    mindist[i] = MAXVALUE;
    relcluster[i] = 0;
    for (j=0; j<nbrel; j++) {
      if (i == relevance[j]) { relcluster[i] = 1; break; }
    }
    mincluster[i] = i;
    for (j=0; j<img_set->num_img; j++) {
      dist = distbuf[row(i, j)][col(i, j)];
      if (mindist[i] > dist && i != j) {
        mindist[i] = dist;
        mincluster[i] = j;
      }
    }
  }
      

  for (i=0; i<img_set->num_img; i++) {
    img_set->img[i].num_img = 1;
    img_set->img[i].next  = NULL;
    img_set->img[i].child = NULL;
    img_set->img[i].key   = NULL;
    img_set->img[i].nearest = NULL;
  }

  if (DEBUG) fprintf(stdout, "Begin Clustering\n");

  key_set->num_img = 0;
  key_set->root=img_set->root=agglomerate2(img_set, key_set);

  for (i=0; i<img_set->num_img; i++) {
    img_set->img[i].size = 0;
  }

  /* Free proximity matrix */
  for (i=0; i<img_set->num_img; i++) free(distbuf[i]);
  free(distbuf);
  free(disttbl);


  return(0);

}

static struct img_info *agglomerate3(struct img_set *img_set, struct img_set *key_set) 
{
  struct img_info *curkey, *nextkey, *newkey;
  float  dist, dist1, dist2, dist3;
  int    num1, num2, num;
  int    i, j, k, ii, jj;
  int    maxcluster, newmax;

  maxcluster = img_set->num_img;
  for (i=img_set->num_img; i>1; i--) {

	if (DEBUG) fprintf(stdout, "Cluster %d\n", i);

    /* Find closest clusters */ 
    dist = MAXVALUE;
    num  = img_set->num_img;
    ii = jj = 0;
    for (j=0; j<maxcluster; j++) {
      if (clusterptr[j] == NULL) continue;
      if (dist > mindist[j]) {
        ii = j; jj = mincluster[j]; dist = mindist[j];
      } else if (dist == mindist[j]) {
        if (((relcluster[ii] == 0 || relcluster[jj] == 0) &&
              relcluster[j] == 1 && relcluster[mincluster[j]] == 1) ||
             (clusterptr[ii]->num_img + clusterptr[jj]->num_img >
              clusterptr[j]->num_img + clusterptr[mincluster[j]]->num_img)) {
          ii = j; jj = mincluster[j]; dist = mindist[j];
        }
      }
    }

    if (DEBUG) fprintf(stdout, "Closest cluster %d %d dist %g\n", ii, jj, mindist[ii]);

    /* create a new cluster with children ii, jj */
    curkey = clusterptr[ii];
    nextkey = clusterptr[jj];
    append_img_info(key_set, curkey);
    newkey = &(key_set->img[key_set->num_img-1]);
    strcpy(newkey->name, "KEY");

    newkey->type      = 1;
    newkey->num_img   = curkey->num_img + nextkey->num_img;
    newkey->dist      = i;
    newkey->size      = mindist[ii];
    newkey->nearest   = NULL;
    newkey->key       = NULL;
    newkey->next      = NULL;
    newkey->child     = curkey;
    curkey->next      = nextkey;
    curkey->key       = newkey;
    nextkey->key      = newkey;

    if (DEBUG) fprintf(stdout, "Updating Proximity matrix\n");
    /* Update Proximity matrix, clusterptr */
    /* replace ii */

    num1  = clusterptr[ii]->num_img; 
    num2  = clusterptr[jj]->num_img; 
    dist3 = newkey->size;
    for (j=0; j<maxcluster; j++) {
      if (clusterptr[j] == NULL) continue;
      dist1 = distbuf[row(ii, j)][col(ii, j)];
      dist2 = distbuf[row(jj, j)][col(jj, j)];

      /* Flexible - Dilating (beta=-1) */
      distbuf[row(ii, j)][col(ii, j)] = dist1 + dist2 - newkey->size;

    }
 
    /* delete jj */
    if (DEBUG) fprintf(stdout, "Delete cluster jj\n");
    clusterptr[ii] = newkey;
    clusterptr[jj] = NULL;
    if (relcluster[ii] == 1 && relcluster[jj] == 1) relcluster[ii] = 1;
    else                                            relcluster[ii] = 0;
    mincluster[jj] = -1;
    mindist[jj] = MAXVALUE;
    for (j=0; j<maxcluster; j++) distbuf[row(jj, j)][col(jj, j)] = MAXVALUE;
    for (newmax=maxcluster-1; newmax>=0; newmax--)
      if (clusterptr[newmax] != NULL) break;
    maxcluster = newmax + 1;


    distbuf[row(ii, ii)][col(ii, ii)] = 0;
    distbuf[row(jj, jj)][col(jj, jj)] = 0;

    /* Update mindist/mincluster table */      
    if (DEBUG) fprintf(stdout, "Updating mindist/mincluster\n");
    for (j=0; j<maxcluster; j++) {
      if (clusterptr[j] == NULL) continue;
      if (mincluster[j] == ii || mincluster[j] == jj) {
        num1 = clusterptr[j]->num_img;
        mindist[j] = MAXVALUE;
        mincluster[j] = 0;
        for (k=0; k<maxcluster; k++) {
          if (clusterptr[k] == NULL) continue;
          dist = distbuf[row(j, k)][col(j, k)];
          if (mindist[j] > dist && k != j) {
            mindist[j] = dist;
            mincluster[j] = k;
          } else if (mindist[j] == dist) {
            if (relcluster[k] == relcluster[mincluster[j]] &&
                clusterptr[k]->num_img < clusterptr[mincluster[j]]->num_img) {
              mindist[j] = dist;
              mincluster[j] = k;
            } else if (relcluster[j] == 1 && 
                       relcluster[k] == 1 && relcluster[mincluster[j]] == 0) {
              mindist[j] = dist;
              mincluster[j] = k;
            }
          }
        }
      }
    }

  }

  return(newkey);

}



