#include "defs.h"

#define INITIAL_ALLOCATION 32



void initialize_node_set(struct node_set *ns)
{
  initialize_heap(&(ns->h));

  ns->N_nodes = 0;
  ns->alloc = INITIAL_ALLOCATION;
  ns->p = (struct node *)mget_spc(ns->alloc,sizeof(struct node));
}


void reset_node_set(struct node_set *ns)
{
  ns->h.size=0;
  ns->N_nodes = 0;
}


void insert_into_node_set(struct node_set *ns, struct node *p, double dist)
{
  int N_nodes;
  struct heap_node hn;
  int i;
  char *pt1,*pt2;

  N_nodes = ns->N_nodes;

  /* if allocation for node array is too small, increase it by *2 */
  if(ns->alloc < N_nodes+2) {
    ns->alloc *=2;
    ns->p = (struct node *)
       realloc((void *)(ns->p),(size_t)(ns->alloc)*sizeof(struct node));
    if(ns->p == NULL) {
      fprintf(stderr,
        "\nError: Cannot allocate for node list, size : %d\n",ns->alloc);
      exit(1);
    }
  }

  /* Copy structure *p to ns->p[N_nodes] */
  pt1 = (char *)(&ns->p[N_nodes]);
  pt2 = (char *)p;
  for(i=0; i<sizeof(struct node); i++) pt1[i] = pt2[i];

  /* Initialize associated heap node */
  hn.x = -dist; /* use negative to look for smallest distance */
  hn.pt = N_nodes;
  (ns->N_nodes)++;

  heap_insert(&(ns->h), &hn);
}


int get_best_from_node_set(struct node_set *ns, struct node *p, double *dist)
{
  struct heap_node hn;
  int i;
  char *pt1,*pt2;

  get_heap_max(&(ns->h), &hn);

  if( hn.pt == -1 ) {
    return(-1);
  }

  *dist = -hn.x;

  /* Copy structure *p to ns->p[N_nodes] */
  pt1 = (char *)p;
  pt2 = (char *)(&ns->p[hn.pt]);
  for(i=0; i<sizeof(struct node); i++) pt1[i] = pt2[i];

  return(1);
}


void free_node_set(struct node_set *ns)
{
  free_heap(&(ns->h));
  free( (void *)(ns->p));
}
